@@ -16,7 +16,6 @@ from dataclasses import dataclass
 from pathlib import Path
 from typing import Any, List, NamedTuple, Sequence, Union, cast
 
-from lsprotocol import validators
 from lsprotocol.types import (
     CODE_ACTION_RESOLVE,
     INITIALIZE,
@@ -31,6 +30,7 @@ from lsprotocol.types import (
     TEXT_DOCUMENT_DID_SAVE,
     TEXT_DOCUMENT_FORMATTING,
     TEXT_DOCUMENT_HOVER,
+    TEXT_DOCUMENT_RANGE_FORMATTING,
     AnnotatedTextEdit,
     ClientCapabilities,
     CodeAction,
@@ -50,6 +50,8 @@ from lsprotocol.types import (
     DidSaveNotebookDocumentParams,
     DidSaveTextDocumentParams,
     DocumentFormattingParams,
+    DocumentRangeFormattingParams,
+    DocumentRangeFormattingRegistrationOptions,
     Hover,
     HoverParams,
     InitializeParams,
@@ -64,13 +66,16 @@ from lsprotocol.types import (
     NotebookDocumentSyncOptionsNotebookSelectorType2CellsType,
     OptionalVersionedTextDocumentIdentifier,
     Position,
+    PositionEncodingKind,
     Range,
     TextDocumentEdit,
+    TextDocumentFilter_Type1,
     TextEdit,
     WorkspaceEdit,
 )
 from packaging.specifiers import SpecifierSet, Version
 from pygls import server, uris, workspace
+from pygls.workspace.position_codec import PositionCodec
 from typing_extensions import Literal, Self, TypedDict, assert_never
 
 from ruff_lsp import __version__, utils
@@ -141,6 +146,7 @@ TOOL_DISPLAY = "Ruff"
 # Require at least Ruff v0.0.291 for formatting, but allow older versions for linting.
 VERSION_REQUIREMENT_FORMATTER = SpecifierSet(">=0.0.291")
 VERSION_REQUIREMENT_LINTER = SpecifierSet(">=0.0.189")
+VERSION_REQUIREMENT_RANGE_FORMATTING = SpecifierSet(">=0.2.1")
 # Version requirement for use of the `--output-format` option
 VERSION_REQUIREMENT_OUTPUT_FORMAT = SpecifierSet(">=0.0.291")
 # Version requirement after which Ruff avoids writing empty output for excluded files.
@@ -1210,14 +1216,47 @@ async def apply_format(arguments: tuple[TextDocument]):
 
 @LSP_SERVER.feature(TEXT_DOCUMENT_FORMATTING)
 async def format_document(params: DocumentFormattingParams) -> list[TextEdit] | None:
+    return await _format_document_impl(params, None)
+
+
+@LSP_SERVER.feature(
+    TEXT_DOCUMENT_RANGE_FORMATTING,
+    DocumentRangeFormattingRegistrationOptions(
+        document_selector=[
+            TextDocumentFilter_Type1(language="python", scheme="file"),
+            TextDocumentFilter_Type1(language="python", scheme="untitled"),
+        ],
+        ranges_support=False,
+        work_done_progress=False,
+    ),
+)
+async def format_document_range(
+    params: DocumentRangeFormattingParams,
+) -> list[TextEdit] | None:
+    return await _format_document_impl(
+        DocumentFormattingParams(
+            params.text_document, params.options, params.work_done_token
+        ),
+        params.range,
+    )
+
+
+async def _format_document_impl(
+    params: DocumentFormattingParams, range: Range | None
+) -> list[TextEdit] | None:
     # For a Jupyter Notebook, this request can only format a single cell as the
     # request itself can only act on a text document. A cell in a Notebook is
     # represented as a text document. The "Notebook: Format notebook" action calls
     # this request for every cell.
     document = Document.from_cell_or_text_uri(params.text_document.uri)
+
     settings = _get_settings_by_document(document.path)
 
-    result = await _run_format_on_document(document, settings)
+    # We don't support range formatting of notebooks yet but VS Code
+    # doesn't seem to respect the document filter. For now, format the entire cell.
+    range = None if document.kind is DocumentKind.Cell else range
+
+    result = await _run_format_on_document(document, settings, range)
     if result is None:
         return None
 
@@ -1376,16 +1415,49 @@ def _fixed_source_to_edits(
         fixed_source = "".join(fixed_source)
 
     new_source = _match_line_endings(original_source, fixed_source)
+
     if new_source == original_source:
         return []
 
+    # Reduce the text edit by omitting the common suffix and postfix (lines)
+    # from the text edit. I chose this basic diffing because "proper" diffing has
+    # the downside that it can be very slow in some cases. Black uses a diffing approach
+    # that takes time into consideration, but it requires spawning a thread to stop
+    # the diffing after a given time, which feels very heavy weight.
+    # This basic "diffing" has a guaranteed `O(n)` runtime and is sufficient to
+    # prevent transmitting the entire source document when formatting a range
+    # or formatting a document where most of the code remains unchanged.
+    #
+    # https://github.com/microsoft/vscode-black-formatter/blob/main/bundled/tool/lsp_edit_utils.py
+    new_lines = new_source.splitlines(True)
+    original_lines = original_source.splitlines(True)
+
+    start_offset = 0
+    end_offset = 0
+
+    for new_line, original_line in zip(new_lines, original_lines):
+        if new_line == original_line:
+            start_offset += 1
+        else:
+            break
+
+    for new_line, original_line in zip(
+        reversed(new_lines[start_offset:]), reversed(original_lines[start_offset:])
+    ):
+        if new_line == original_line:
+            end_offset += 1
+        else:
+            break
+
+    trimmed_new_source = "".join(new_lines[start_offset : len(new_lines) - end_offset])
+
     return [
         TextEdit(
             range=Range(
-                start=Position(line=0, character=0),
-                end=Position(line=validators.UINTEGER_MAX_VALUE, character=0),
+                start=Position(line=start_offset, character=0),
+                end=Position(line=len(original_lines) - end_offset, character=0),
             ),
-            new_text=new_source,
+            new_text=trimmed_new_source,
         )
     ]
 
@@ -1824,14 +1896,19 @@ async def _run_check_on_document(
 
 
 async def _run_format_on_document(
-    document: Document, settings: WorkspaceSettings
+    document: Document, settings: WorkspaceSettings, format_range: Range | None = None
 ) -> ExecutableResult | None:
     """Runs the Ruff `format` subcommand on the given document source."""
     if settings.get("ignoreStandardLibrary", True) and document.is_stdlib_file():
         log_warning(f"Skipping standard library file: {document.path}")
         return None
 
-    executable = _find_ruff_binary(settings, VERSION_REQUIREMENT_FORMATTER)
+    version_requirement = (
+        VERSION_REQUIREMENT_FORMATTER
+        if format_range is None
+        else VERSION_REQUIREMENT_RANGE_FORMATTING
+    )
+    executable = _find_ruff_binary(settings, version_requirement)
     argv: list[str] = [
         "format",
         "--force-exclude",
@@ -1840,6 +1917,19 @@ async def _run_format_on_document(
         document.path,
     ]
 
+    if format_range:
+        codec = PositionCodec(PositionEncodingKind.Utf16)
+        format_range = codec.range_from_client_units(
+            document.source.splitlines(True), format_range
+        )
+
+        argv.extend(
+            [
+                "--range",
+                f"{format_range.start.line + 1}:{format_range.start.character + 1}-{format_range.end.line + 1}:{format_range.end.character + 1}",  # noqa: E501
+            ]
+        )
+
     for arg in settings.get("format", {}).get("args", []):
         if arg in UNSUPPORTED_FORMAT_ARGS:
             log_to_output(f"Ignoring unsupported argument: {arg}")
